import os

import jax

jax.config.update("jax_enable_x64", True)
os.environ["XLA_FLAGS"] = f"--xla_force_host_platform_device_count={os.cpu_count()}"

import numpy as np
from tinygp import GaussianProcess, kernels

from nuance import core, linear_search, periodic_search

depth = 2.3e-3
error = 1e-3
time = np.linspace(0, 3.0, 4000)
transit_params = {"epoch": 0.2, "duration": 0.05, "period": 0.7}
transit_model = depth * core.transit(time, **transit_params)
kernel = kernels.quasisep.SHO(10.0, 10.0, 0.002)
gp = GaussianProcess(kernel, time, diag=error**2)
flux = transit_model + gp.sample(jax.random.PRNGKey(0)) + 1.0


def test_full():
    # linear search
    epochs = time.copy()
    durations = np.linspace(0.01, 0.2, 15)
    ls = linear_search(time, flux, gp=gp)(epochs, durations)

    # periodic search
    periods = np.linspace(0.3, 5, 2000)
    snr_function = jax.jit(core.snr(time, flux, gp=gp))
    ps_function = periodic_search(epochs, durations, ls, snr_function)
    snr, params = ps_function(periods)

    period = params[np.argmax(snr)][-1]
    np.testing.assert_allclose(period, transit_params["period"], atol=0.001)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = "nuance"
copyright = "2023, Lionel Garcia"
author = "Lionel Garcia"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "myst_nb",
    "sphinx_copybutton",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosummary",
    "sphinx_design",
]


templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build/*"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_book_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Title
# get version number from pyproject.toml
# --------------------------------------
# import toml

# pyproject = toml.load("../pyproject.toml")
# version = pyproject["tool"]["poetry"]["version"]
# html_short_title = "nuance"
# html_title = f"{html_short_title}"
html_logo = "_static/logo.png"
html_favicon = "_static/favicon.png"
# -----

source_suffix = {
    ".rst": "restructuredtext",
    ".ipynb": "myst-nb",
    ".myst": "myst-nb",
}

root_doc = "index"

html_theme_options = {
    "repository_url": "https://github.com/lgrcia/nuance",
    "use_repository_button": True,
    "use_edit_page_button": True,
    "use_issues_button": True,
    "default_mode": "light",
}

nb_render_image_options = {"align": "center"}

html_css_files = ["style.css"]

myst_enable_extensions = [
    "dollarmath",
]

autodoc_typehints = "signature"
autoclass_content = "both"

nb_execution_excludepatterns = [
    "_build/*",
    "notebooks/tutorials/*",
    "notebooks/_*.ipynb",
]



import pytest


@pytest.mark.parametrize("cpu_count", [None, 1, 2, 4])
def test_cpu(cpu_count):
    pytest.skip("Figure out later")
    import os

    import jax

    if cpu_count is None:
        cpu_count = os.cpu_count()

    os.environ["XLA_FLAGS"] = f"--xla_force_host_platform_device_count={cpu_count}"

    import nuance

    assert jax.device_count() == cpu_count



import numpy as np
from astropy import constants as c
from astropy import units as u

from nuance.star import Star

R = 1.0 * u.R_sun
M = 1.0 * u.M_sun
star = Star(R.value, M.value, 0, 0)


def test_duration():
    period = 3.0 * u.day

    a = (c.G * M * (period**2) / (4 * (np.pi**2))) ** (1 / 3)
    duration = (period * R) / (np.pi * a)
    expected = duration.to(u.day).value

    star = Star(R.value, M.value, 0, 0)
    computed = star.transit_duration(period.value)

    np.testing.assert_allclose(expected, computed)


def test_depth():
    radius = 1.0 * u.R_earth
    expected = ((radius / R) ** 2).decompose().value
    computed = star.transit_depth(radius.value)
    np.testing.assert_allclose(expected, computed)



import jax
import jaxopt
import numpy as np
import tinygp
from scipy.ndimage import minimum_filter1d

from nuance import core


def interp_split_times(time, p, dphi=0.01):
    tmax, tmin = np.max(time), np.min(time)
    total = tmax - tmin
    # since for very small periods we might fold on few points only, it's better to impose
    # at least 200 phases, to compare period folds more fairly
    phase = np.arange(0, 1, np.min([1 / 200, dphi / p]))
    n = np.arange(np.ceil(total / p) + 1)  # number of 'folds'
    # this line is important so that different time-series have the same phase 0
    tmin -= tmin % p
    pt0s = np.array([tmin + phase * p + j * p for j in n])  # corresponding t0s
    has_time = np.any([np.abs(time - _t0) < p / 2 for _t0 in pt0s.mean(1)], 1)
    pt0s = pt0s[has_time]

    return pt0s


def time_phase_resample(time, phase, period):
    t0 = np.min(time)
    duration = np.ptp(time)
    n = np.arange(np.ceil(duration / period) + 1)  # number of 'folds'
    t0 -= t0 % period  # phase 0
    t0s = np.array([t0 + (phase + j) * period for j in n])  # corresponding t0s
    has_time = np.any([np.abs(time - _t0) < period / 2 for _t0 in t0s.mean(1)], 1)
    return t0s[has_time]


def phase(t, t0, p):
    return (t - t0 + 0.5 * p) % p - 0.5 * p


def log_gauss_product_integral(a, va, b, vb):
    """Log of the product ot two Normal distributions (normalized)

    Parameters
    ----------
    a : float
        mean of distribution A
    va : float
        variance of distribution A
    b : float
        mean of distribution B
    vb : float
        variance of distribution B

    Returns
    -------
    float
        log of the integral of A*B
    """
    return (
        -0.5 * np.square(a - b) / (va + vb)
        - 0.5 * np.log(va + vb)
        - 0.5 * np.log(np.pi)
        - np.log(2) / 2
    )


def simulated(
    t0=0.2, D=0.05, depth=0.02, P=0.7, time=None, kernel=None, error=0.001, w=None
):
    if time is None:
        time = np.arange(0, 4, 2 / 60 / 24)

    X = np.vander(time, N=4, increasing=True).T
    if w is None:
        w = [1.0, 5e-4, -2e-4, -5e-4]

    true_transit = depth * core.transit(time, t0, D, period=P)

    if kernel is None:
        kernel = tinygp.kernels.quasisep.SHO(np.pi / (6 * D), 45.0, depth)

    gp = tinygp.GaussianProcess(kernel, time, diag=error**2, mean=0.0)
    flux = gp.sample(jax.random.PRNGKey(40)) + true_transit + w @ X

    return (time, flux, error), X, gp


def clean_periods(periods, period, tol=0.02):
    close = periods / period
    return periods[np.abs(close - np.rint(close)) > tol]


def index_binning(x, size):
    if isinstance(size, float):
        bins = np.arange(np.min(x), np.max(x), size)
    else:
        x = np.arange(0, len(x))
        bins = np.arange(0.0, len(x), size)

    d = np.digitize(x, bins)
    n = np.max(d) + 2
    indexes = []

    for i in range(0, n):
        s = np.where(d == i)
        if len(s[0]) > 0:
            s = s[0]
            indexes.append(s)

    return indexes


def binn_time(time, flux, bins=7 / 24 / 60):
    indexes = index_binning(time, bins)
    binned_time = np.array([np.mean(time[i]) for i in indexes])
    binned_flux = np.array([np.mean(flux[i]) for i in indexes])
    binned_error = np.array([np.std(flux[i]) / np.sqrt(len(i)) for i in indexes])
    return binned_time, binned_flux, binned_error


def minimize(fun, init_params, param_names=None):
    """Minimize a function using jaxopt.ScipyMinimize

    Parameters
    ----------
    fun : callable
        the function to minimize of the form fun(params: dict) -> float
    init_params : dict
        initial parameters
    param_names : list, optional
        list of parameters to optimize, all others being fixed,
        by default None

    Returns
    -------
    dict
        optimized parameters
    """

    def inner(theta, *args, **kwargs):
        params = dict(init_params, **theta)
        return fun(params, *args, **kwargs)

    param_names = list(init_params.keys()) if param_names is None else param_names
    start = {k: init_params[k] for k in param_names}

    solver = jaxopt.ScipyMinimize(fun=inner)
    soln = solver.run(start)

    return dict(init_params, **soln.params)


def sigma_clip_mask(residuals, sigma=5, window=20):
    """Returns a masked of the sigma clipped data. The mask is set
        to false on the +/-{windows} data points around the sigma
        clipped residuals, i.e. code:`residuals > sigma * std(residuals)`.

    Parameters
    ----------
    residuals : array
        Data to sigma clip.
    sigma : int, optional
        Sigma clipping standard deviation, by default 5.
    window : int, optional
        The window of points around which mask is set to False if
        a data point is sigma clipped, by default 20

    Returns
    -------
    array
        boolean mask of sigma clipped values (i.e :code:`<std(residuals)`).
    """
    mask = np.ones_like(residuals, dtype=bool)
    mask[residuals > sigma * np.std(residuals)] = False
    mask = np.roll(minimum_filter1d(mask, window), shift=window // 3)
    return mask



import jax
import jax.numpy as jnp
import numpy as np
from tqdm.auto import tqdm
from typing import Callable
import tinygp

from nuance import DEVICES_COUNT, core


def linear_search(
    time: np.ndarray,
    flux: np.ndarray,
    gp: tinygp.GaussianProcess | None = None,
    X: np.ndarray | None = None,
    model: Callable | None = None,
    positive: bool = True,
    progress: bool = True,
    backend: str | None = None,
    batch_size: int | None = None,
):
    """Returns a function that computes the log likelihood of a transit model at
    different epochs and durations (linear search)

    Parameters
    ----------
    time : np.ndarray
        array of times
    flux : np.ndarray
        flux time-series
    gp : tinygp.GaussianProcess, optional
        tinygp GaussianProcess model, by default None
    X : np.ndarray, optional
        linear model design matrix, by default None
    positive : bool, optional
        wether to force depth to be positive, by default True
    progress : bool, optional
        wether to show progress bar, by default True
    backend : str, optional
        backend to use, by default jax.default_backend() (options: "cpu", "gpu").
        This affects the linear search function jax-mapping strategy. For more details, see
        :py:func:`nuance.core.map_function`
    batch_size : int, optional
        batch size for parallel evaluation, by default None
    """
    X, gp, model = core.check_default(time, X, gp, model)

    solver = jax.jit(core.solve(time, flux, gp=gp, X=X))

    if backend is None:
        backend = jax.default_backend()

    if batch_size is None:
        batch_size = {"cpu": DEVICES_COUNT, "gpu": 1000}[backend]

    if backend == "cpu":
        solve_batch = jax.pmap(jax.vmap(solver, in_axes=(None, 0)), in_axes=(0, None))
    else:
        solve_batch = jax.jit(
            jax.vmap(jax.vmap(solver, in_axes=(None, 0)), in_axes=(0, None))
        )

    def function(epochs: np.ndarray, durations: np.ndarray):
        """Compute the log likelihood of a transit model at different epochs and durations

        Parameters
        ----------
        epochs : np.ndarray
            array of epochs
        durations : np.ndarray
            array of durations

        Returns
        -------
        tuple
            (log likelihoods, model depths, depths variances)
        """
        epochs_padded = np.pad(
            epochs, [0, batch_size - (len(epochs) % batch_size) % batch_size]
        )
        epochs_batches = np.reshape(
            epochs_padded, (len(epochs_padded) // batch_size, batch_size)
        )

        _progress = lambda x: (tqdm(x, unit_scale=batch_size) if progress else x)

        results = []

        for epoch_batch in _progress(epochs_batches):
            results.append(solve_batch(epoch_batch, durations))

        depths, vars, log_likelihoods = np.transpose(
            results, axes=[3, 0, 1, 2]
        ).reshape((3, len(epochs_padded), len(durations)))[:, 0 : len(epochs), :]

        if positive:
            ll0 = core.solve_model(flux, X, gp)(jnp.zeros_like(time))[0]
            log_likelihoods[depths < 0] = ll0

        vars[~np.isfinite(vars)] = 1e25

        # trick to store ll0
        log_likelihoods[0, 0] = ll0
        depths[0, 0] = 0.0

        return log_likelihoods, depths, vars

    return function


def combine_linear_searches(*linear_searches):
    """Combine the results of multiple linear searches

    *linear_searches : list
        lists of (log likelihoods, model depths, depths variances)

    Returns
    -------
    tuple
        (log likelihoods, model depths, depths variances)

    Example
    -------
    .. code-block:: python

        ls, z, vz = combine_linear_searches((ls0, z0, vz0), (ls1, z1, vz1), (ls2, z2, vz2))

    """
    ll = np.vstack([ls[0] for ls in linear_searches])
    z = np.vstack([ls[1] for ls in linear_searches])
    vz = np.vstack([ls[2] for ls in linear_searches])
    ll0s = [ls[0][0, 0] for ls in linear_searches]
    ll = (
        ll
        - np.hstack(
            [
                np.ones(linear_searches[i][0].shape[0]) * ll0s[i]
                for i in range(len(linear_searches))
            ]
        )[:, None]
    ) + np.sum(ll0s)

    return np.array([ll, z, vz])



import multiprocess as mp
import numpy as np
from tqdm.auto import tqdm

from nuance import core
from nuance.utils import interp_split_times


def periodic_search(epochs, durations, ls, snr_f, progress=True):
    """Returns a function that performs the periodic search given an array of periods.

    Parameters
    ----------
    epochs : np.ndarray
        Values of epochs for with the likelihoods have been computed.
    durations : np.ndarray
        Values of durations for with the likelihoods have been computed.
    ls : list, tuple, np.ndarray
        Tuple of (likelihoods, depth, depths variance).
    snr_f : callable
        Function that computes the SNR given the epoch, duration and period.
    progress : bool, optional
        wether to show progress bar, by default True

    Returns
    -------
    callable
        Function that computes the SNR and parameters for each period.
    """
    global fold_f
    fold_f = _fold_ll(epochs, *ls)

    def _progress(x, **kwargs):
        return tqdm(x, **kwargs) if progress else x

    def function(periods):
        snr = np.zeros(len(periods))
        params = np.zeros((len(periods), 3))

        with mp.Pool() as pool:
            for p, (epoch, duration_i, period) in enumerate(
                _progress(pool.imap(_solve, periods), total=len(periods))
            ):
                Dj = durations[duration_i]
                snr[p], params[p] = float(snr_f(epoch, Dj, period)), (epoch, Dj, period)

        return snr, params

    return function


def _fold_ll(epochs, lls, z, vz):
    f_ll = core.nearest_neighbors(epochs, lls)
    f_z = core.nearest_neighbors(epochs, z)
    f_dz2 = core.nearest_neighbors(epochs, vz)

    def _fold(times):
        lls = np.array([f_ll(time) for time in times])
        zs = np.array([f_z(time) for time in times])
        vzs = np.array([f_dz2(time) for time in times])

        P1 = np.sum(lls, 0)
        vZ = 1 / np.sum(1 / vzs, 0)
        Z = vZ * np.sum(zs / vzs, 0)
        P1 = np.sum(lls, 0)
        P2 = 0.5 * np.sum(
            np.log(vzs) - np.log(vzs + vZ) + (zs - Z) ** 2 / (vzs + vZ), 0
        )

        return P1 - P2

    def fun(period):
        times = interp_split_times(epochs, period)
        phase = times[0] / period
        lls = _fold(times)
        return phase, lls

    return fun


def _solve(period):
    phase, lls = fold_f(period)
    epoch_i, duration_i = np.unravel_index(np.argmax(lls), lls.shape)
    epoch = phase[epoch_i] * period
    return epoch, duration_i, period



import jax
import jax.numpy as jnp
from jax.scipy.linalg import block_diag
from tinygp import GaussianProcess, kernels

INPUTS_DOCS = """
    time : np.ndarray
        array of times
    flux : np.ndarray
        array of fluxes
    X : np.ndarray, optional
        linear model design matrix, by default a constant model
    gp : tinygp.GaussianProcess, optional
        Gaussian process object, by default a very long scale exponential kernel
    model : callable, optional
        model function with signature :code:`model(time, epoch, duration, period=None) -> Array`,
        by default an empirical :py:func:`~nuance.core.transit` model
"""

DEFAULT_X = lambda time: jnp.atleast_2d(jnp.ones_like(time))
DEFAULT_GP = lambda time: GaussianProcess(kernels.quasisep.Exp(1e12), time)


def check_default(time, X=None, gp=None, model=None):
    if X is None:
        X = DEFAULT_X(time)
    if gp is None:
        gp = DEFAULT_GP(time)
    if model is None:
        model = transit
    return X, gp, model


def solve_triangular(*gps_y):
    Ls = [gp.solver.solve_triangular(y) for gp, y in gps_y]
    if Ls[0].ndim == 1:
        return jnp.hstack(Ls)
    else:
        return block_diag(*Ls)


def solve_model(flux, X, gp):

    # multiple gps and datasets
    if isinstance(flux, (list, tuple)):
        assert (
            len(gp) == len(flux) == len(X)
        ), "gp, flux, and datasets must have the same length"
        Liy = solve_triangular(*[(_gp, _flux) for _gp, _flux in zip(gp, flux)])
        LiX = solve_triangular(*[(_gp, _X.T) for _gp, _X in zip(gp, X)])

        @jax.jit
        def function(ms):
            Lim = solve_triangular(*[(_gp, m) for _gp, m in zip(gp, ms)])
            LiXm = jnp.hstack([LiX, Lim[:, None]])
            LiXmT = LiXm.T
            LimX2 = LiXmT @ LiXm
            w = jnp.linalg.lstsq(LimX2, LiXmT @ Liy)[0]
            v = jnp.linalg.inv(LimX2)
            return 0.0, w, v

        return function

    # single gp and dataset
    else:
        Liy = gp.solver.solve_triangular(flux)
        LiX = gp.solver.solve_triangular(X.T)

        @jax.jit
        def function(m):
            Xm = jnp.vstack([X, m])
            Lim = gp.solver.solve_triangular(m)
            LiXm = jnp.hstack([LiX, Lim[:, None]])
            LiXmT = LiXm.T
            LimX2 = LiXmT @ LiXm
            w = jnp.linalg.lstsq(LimX2, LiXmT @ Liy)[0]
            v = jnp.linalg.inv(LimX2)
            return gp.log_probability(flux - w @ Xm), w, v

        return function


def solve(time, flux, gp=None, X=None, model=None):
    """Returns a function to compute the log likelihood of data assuming it is drawn
    from a Gaussian Process with a mean linear model.

    `X` is a design matrix for the linear model whose last column is the searched signal
    :code:`model`.

    Parameters
    ----------
    time : np.ndarray
        array of times
    flux : np.ndarray
        array of fluxes
    X : np.ndarray, optional
        linear model design matrix, by default a constant model
    gp : tinygp.GaussianProcess, optional
        Gaussian process object, by default a very long scale exponential kernel
    model : callable, optional
        model function with signature :code:`model(time, epoch, duration, period=None) -> Array`,
        by default an empirical :py:func:`~nuance.core.transit` model

    Returns
    -------
    callable
        function that computes the log likelihood of data assuming it is drawn from a
        Gaussian Process with a mean linear model. Signature is:
        :code:`function(epoch, duration, period=None) -> (log_likelihood, weights, variance)`
    """

    X, gp, model = check_default(time, X, gp, model)

    if model is None:
        model = transit

    solve_m = solve_model(flux, X, gp)

    # multiple datasets
    if isinstance(time, (list, tuple)):

        def _model(time, epoch, duration, period=None):
            return [model(t, epoch, duration, period=period) for t in time]

    else:
        _model = model

    def function(epoch, duration, period=None):
        m = _model(time, epoch, duration, period=period)
        ll, w, v = solve_m(m)
        return jnp.array([w[-1], v[-1, -1], ll])

    return function


def gp_model(x, y, build_gp, X=None):

    if X is None:
        X = jnp.atleast_2d(jnp.ones_like(x))

    @jax.jit
    def nll_w(params):
        gp = build_gp(params, x)
        Liy = gp.solver.solve_triangular(y)
        LiX = gp.solver.solve_triangular(X.T)
        LiXT = LiX.T
        LiX2 = LiXT @ LiX
        w = jnp.linalg.lstsq(LiX2, LiXT @ Liy)[0]
        nll = -gp.log_probability(y - w @ X)
        return nll, w

    @jax.jit
    def nll(params):
        return nll_w(params)[0]

    @jax.jit
    def mu(params):
        gp = build_gp(params, x)
        _, w = nll_w(params)
        cond_gp = gp.condition(y - w @ X, x).gp
        return cond_gp.loc + w @ X

    return mu, nll


def transit(time, epoch, duration, period=None, c=12):
    """Empirical transit model from Protopapas et al. 2005.

    Parameters
    ----------
    time : np.ndarray
        array of times
    epoch : float
        signal epoch
    duration : float
        signal duration
    period : float, optional
        signal period, by default None for a non-periodic signal
    c : int, optional
        controls the 'roundness' of transit shape, by default 12

    Returns
    -------
    np.ndarray
        array of transit model values
    """
    if period is None:
        period = 1e15
    _t = period * jnp.sin(jnp.pi * (time - epoch) / period) / (jnp.pi * duration)
    return -0.5 * jnp.tanh(c * (_t + 1 / 2)) + 0.5 * jnp.tanh(c * (_t - 1 / 2))


def transit_box(time, epoch, duration, period=1e15):
    """Box-shaped transit model.

    Parameters
    ----------
    time : np.ndarray
        array of times
    epoch : float
        signal epoch
    duration : float
        signal duration
    period : float, optional
        signal period, by default None for a non-periodic signal

    Returns
    -------
    np.ndarray
        array of transit model values
    """
    return -((jnp.abs(time - epoch) % period) < duration / 2).astype(float)


def transit_exocomet(time, epoch, duration, period=None, n=3):
    """Empirical exocomet transit model.

    Parameters
    ----------
    time : np.ndarray
        array of times
    epoch : float
        signal epoch
    duration : float
        signal duration
    period : float, optional
        dummy parameter for compatibility with other models, by default None
    n : int, optional
        TBD, by default 3

    Returns
    -------
    np.ndarray
        array of transit model values
    """
    flat = jnp.zeros_like(time)
    left = -(time - (epoch - duration / n)) / (duration / n)
    right = -jnp.exp(-2 / duration * (time - epoch - duration / n)) ** 2
    triangle = jnp.maximum(left, right)
    mask = time >= epoch - duration / n
    signal = jnp.where(mask, triangle, flat)
    return signal / jnp.max(jnp.array([-jnp.min(signal), 1]))


def separate_models(time, flux, X=None, gp=None, model=None):
    """Returns a function to compute the mean, signal, and noise model of a light curve
    given the epoch, duration, and period of the signal model.

    Parameters
    ----------
    time : np.ndarray
        array of times
    flux : np.ndarray
        array of fluxes
    X : np.ndarray, optional
        linear model design matrix, by default a constant model
    gp : tinygp.GaussianProcess, optional
        Gaussian process object, by default a very long scale exponential kernel
    model : callable, optional
        model function with signature :code:`model(time, epoch, duration, period=None) -> Array`,
        by default an empirical :py:func:`~nuance.core.transit` model

    Returns
    -------
    callable
        function that computes the mean, signal, and noise model of a light curve given
        the epoch, duration, and period of the signal model. Signature is:
        :code:`function(epoch, duration, period=None) -> (mean, signal, noise)`
    """

    X, gp, model = check_default(time, X, gp, model)

    solver = solve_model(flux, X, gp)

    def function(epoch, duration, period=None):
        m = model(time, epoch, duration, period=period)
        w = solver(m)[1]
        mean = w[0:-1] @ X
        signal = m * w[-1]

        @jax.jit
        def gp_mean():
            return gp.condition(flux - mean - signal).gp.mean

        noise = gp_mean()

        return mean, signal, noise

    return function


def snr(time, flux, X=None, gp=None, model=None):
    """Returns a function to compute the signal-to-noise ratio of a signal model given
    its epoch, duration, and period.

    Parameters
    ----------
    time : np.ndarray
        array of times
    flux : np.ndarray
        array of fluxes
    X : np.ndarray, optional
        linear model design matrix, by default a constant model
    gp : tinygp.GaussianProcess, optional
        Gaussian process object, by default a very long scale exponential kernel
    model : callable, optional
        model function with signature :code:`model(time, epoch, duration, period=None) -> Array`,
        by default an empirical :py:func:`~nuance.core.transit` model

    Returns
    -------
    callable
        function that computes the signal-to-noise ratio of a signal model given its
        epoch, duration, and period. Signature is:
        :code:`function(epoch, duration, period=None) -> float`
    """

    X, gp, model = check_default(time, X, gp, model)

    solver = solve(time, flux, gp=gp, X=X, model=model)

    def function(epoch, duration, period=None):
        z, vz, _ = solver(epoch, duration, period)
        return jnp.max(jnp.array([0, z / jnp.sqrt(vz)]))

    return function


def interpolate(xp, fp):
    import numpy as np

    xp = np.asarray(xp)
    fp = np.asarray(fp.T)

    def fun(x):
        x = np.asarray(x)
        j = np.searchsorted(xp, x) - 1
        d = np.where(j + 1 < len(xp), (x - xp[j]) / (xp[j + 1] - xp[j]), 0)
        return ((1 - d) * fp[:, j] + fp[:, j + 1] * d).T

    return fun


def nearest_neighbors(xp, fp):
    import numpy as np

    xp = np.asarray(xp)
    fp = np.asarray(fp)

    def fun(x):
        x = np.asarray(x)
        j = np.searchsorted(xp, x) - 1
        return fp[j]

    return fun



from dataclasses import dataclass

import numpy as np

G = 2942.2062175044193
# R_sun^3 . M_sun^-1 . day^-2
# astropy: c.G.to(u.R_sun**3/u.M_sun/(u.day**2)).value

R_EARTH = 0.009167888457668534
# R_sun
# astropy: c.R_earth.to(u.R_sun).value


@dataclass
class Star:
    """A class to hold stellar parameters and variability characteristics."""

    radius: float = 1.0
    """Stellar radius in :math:`R_\odot`"""
    mass: float = 1.0
    """Stellar mass in :math:`M_\odot`"""
    amplitude: float = 1.0  # peak to peak
    """Stellar variability amplitude (peak to peak)"""
    period: float = 1.0
    """Stellar variability period in days"""

    def __post_init__(self):
        pass

    @property
    def omega(self) -> float:
        """Stellar variability angular frequency.

        Returns
        -------
        float
            Angular frequency in rad.s-1
        """
        return 2 * np.pi / self.period

    def transit_duration(self, orbital_period: float) -> float:
        """Transit duration from orbital period (days).

        Parameters
        ----------
        orbital_period : float
            planet orbital period in days

        Returns
        -------
        float
            duration in days
        """
        a = (G * self.mass * (orbital_period**2) / (4 * (np.pi**2))) ** (1 / 3)
        return (orbital_period * self.radius) / (np.pi * a)

    def transit_depth(self, radius: float) -> float:
        """Transit depth from planet's radius (in :math:`R_\oplus`).

        Parameters
        ----------
        radius : float
            planet's radius in Earth radius

        Returns
        -------
        float
            transit depth
        """
        _radius = radius * R_EARTH
        return (_radius / self.radius) ** 2

    # period - tau

    def period2tau(self, orbital_period: float) -> float:
        """Relative duration given orbital period.

        Parameters
        ----------
        orbital_period : float
            planet orbital period in days

        Returns
        -------
        float
            relative duration tau
        """
        duration = self.transit_duration(orbital_period)
        return np.pi / (self.omega * duration)

    def tau2period(self, tau: float) -> float:
        """Period given relative duration (in days).

        Parameters
        ----------
        tau : float
            relative duration

        Returns
        -------
        float
            orbital period in days
        """
        duration = np.pi / (self.omega * tau)
        a = (G * self.mass * duration**2) / (4 * (self.radius**2))
        return self._period(a)

    # radius - delta

    def radius2delta(self, radius: float) -> float:
        """Relative amplitude given planet's radius.

        Parameters
        ----------
        radius : float
            planet's radius in Earth radius

        Returns
        -------
        float
            relative amplitude
        """
        depth = self.transit_depth(radius)
        return self.amplitude / depth

    def delta2radius(self, delta: float) -> float:
        """Planet's radius from relative amplitude (in :math:`R_\oplus`).

        Parameters
        ----------
        delta : float
            relative amplitude

        Returns
        -------
        float
            planet's radius in Earth radius
        """
        depth = self.delta2depth(delta)
        return np.sqrt(depth) * self.radius / R_EARTH

    def min_radius(self, period: float, SNR: float, N: int, sigma: float) -> float:
        """Given a target SNR and periods, returns the minimum planetary radius detectable (in :math:`R_\oplus`).

        Parameters
        ----------
        period : float
            planet orbital period in days
        SNR : float
            signal-to-noise-ratio
        N : int
            number of points observed
        sigma : float
            observation error

        Returns
        -------
        float
            radius is earth radius
        """
        D = self.transit_duration(period)
        n = D * N / period
        return (np.sqrt(SNR * sigma) * self.radius / n ** (1 / 4)) / R_EARTH

    def snr(self, orbital_period: float, radius: float, N: int, sigma: float) -> float:
        """Transit signal-to-noise-ratio.

        Parameters
        ----------
        orbital_period : float
            planet orbital period in days
        radius : float
            planet's radius in Earth radius
        N : int
            number of observation points
        sigma : float
            observation error

        Returns
        -------
        float
            signal-to-noise
        """
        depth = self.transit_depth(radius)
        duration = self.transit_duration(orbital_period)
        n_tr = N * duration / orbital_period  # points in transit
        return (depth / sigma) * np.sqrt(n_tr)

    def _period(self, a: float):
        return 2 * np.pi * (a ** (3 / 2)) / np.sqrt(G * self.mass)

    @property
    def density(self) -> float:
        """Stellar density in :math:`M_\odot.R_\odot^{-3}`.

        Returns
        -------
        float
            stellar density
        """
        return self.mass / (4 / 3 * np.pi * self.radius**3)

    def roche_period(self, density: float = 1.0) -> float:
        """Roche limit for a given density and assuming circular orbit (in days).

        Parameters
        ----------
        density : float
            planet density in g.cm-3, by default 1

        Returns
        -------
        float
            Period of a planet in a circular orbit at the Roche limit (days)
        """
        # conversion from
        # import astropy.units as u
        # (u.g/u.cm**3).to(u.solMass/u.R_sun**3)
        density_M_sun_Rsun3 = density * 0.16934021222434983
        a = 2.44 * self.radius * (self.density / density_M_sun_Rsun3) ** (1 / 3)
        return self._period(a)

    def period_grid(self, time_span, period_max=None, period_min=None, oversampling=1):
        """Grid of optimal periods following Ofir (2014) (in days).

        Parameters
        ----------
        time_span : float
            time span in days.
        period_max : float
            maximum period in days, by default None which
            defaults to half the time span.
        period_min : float, optional
            minimum period in days, by default None which
            defaults to the roche limit period.
        oversampling: int, optional
            oversampling factor, by default 1.

        Returns
        -------
        np.ndarray
            periods in days.
        """
        if period_min is None:
            period_min = self.roche_period()
        if period_max is None:
            period_max = time_span / 2

        A = (
            (2 * np.pi) ** (2 / 3)
            / np.pi
            * self.radius
            / (G * self.mass) ** (1 / 3)
            * 1
            / (time_span * oversampling)
        )

        N = 3 * ((1 / period_min) ** (1 / 3) - (1 / period_max) ** (1 / 3)) / A + 1
        frequencies = np.linspace(1 / period_min, 1 / period_max, int(N))
        return 1 / frequencies



import jax

DEVICES_COUNT = jax.device_count()

config = jax.config
config.update("jax_enable_x64", True)

from nuance.linear_search import linear_search
from nuance.periodic_search import periodic_search
from nuance.star import Star
